;! This is a minimalist beat-em-up game written by amul for
;! the first Minimalist Compo on the German Blitz Basic Portal
;! at http://www.blitzforum.de/forum/viewtopic.php?t=39245

;! Also visit http://amulware.net for more games
;! and other stuff by amul.
;! Contact at amul@amulware.net

;! The official download of this file(including graphics and
;! an executeable) is:
;! http://amulware.net/pgr/bb/minibeatemup.zip

;! ===========================================================
;! The controls for this game are as follows:

;! start/game-over screen:
;! [U]: play human vs human
;! [I]: play human vs AI (player 2 is AI)

;! player 1 (arrows and enter):
;! [left arrow]:  close in on opponent
;! [right arrow]: block
;! [down arrow]:  duck
;! [up arrow]:    jump
;! [enter]:       punch

;! player 2 (WASD and left control):
;! [D]: close in on opponent
;! [A]: block
;! [S]: duck
;! [W]: jump
;! [left control]: punch

;! while playing you can also exit the game with the ESC key

;! ===========================================================
;! The code below is commented as follows:

;! Every line includes a comment with it's cost as layed out
;! in the contest's thread. The entire program had to have
;! a cost of at most 256. This program uses 254.
;! Also note that 4096 bytes of image data was allowed, again
;! calcualted as specified in the mentioned thread.
;! This game uses exactly 4091.125 bytes by that measure.

;! Sometimes these comments also include a note on why a certain
;! line has a specific cost.
;! All other comments explain the logic of the code, and tricks
;! used to shorten it.

;! Enjoy reading, and of course playing!

; two graphics calls to make the window big, but the resolution
; small, for great justice and retro style
Graphics(640, 400, 0, 2) ; 0 ; setting up resolution is free
Graphics(160, 100, 0, 3) ; 0

; creating two players(the numbers are the key codes used to
; control the players, except for the last, which is game
; math related)
Global player1.TPlayer = CreatePlayer(203, 205, 28, 208, 200, "gfx/bot.png", 1) ; 9
Global player2.TPlayer = CreatePlayer(32, 30, 29, 31, 17, "gfx/bot2.png", -1) ; 9

; loading images for sky, background, and game over message
Local skyImage = LoadImage("gfx/sky.png") ; 3
Local backgroundImage = LoadImage("gfx/back.png") ; 3
Local gameoverImage = LoadImage("gfx/go.png") ; 3

; initialising well named and semi self explaining globals
Global backgroundParallax ; 0 ; initialisation is free
Global playersIntersect ; 0
Global playersConnect ; 0


; setting green colour for health bars
Color(0, 200, 0) ; 4

; creating timer to keep game running at 20 FPS
Local timer = CreateTimer(20) ; 3 ; timers are expensive, but
                                  ; needed, because Flip(True)
                                  ; often uses 100% CPU

; setting up double buffering
SetBuffer(BackBuffer()) ; 0 ; setting backbuffer once is free

; this loop makes the game reset itself each time a round ends
Repeat ; 1 ; not main loop, thus it costs
	
	; draw game over image, asking player to select human/ai
	; as opponent
	DrawImage(gameoverImage, 43,7) ; 4
	Flip() ; 1 ; not main loop flip costs
	
	; this loop waits for the player to hit a key
	; if it is a U or I, the right opponent is selected
	; otherwise it continues waiting
	Repeat ; 1 ; still no main loop
		; no need to run this loop at 100% CPU, thus
		; waiting for keystroke here
		WaitKey() ; 1
		
		; avoiding branching(Ifs) is a good way to save
		; instructions
		; the line below is (virtually) equivalent to:
		; player2\isAI = -1
		; If KeyHit(22) Then player2\isAI = 0
		; If KeyHit(23) Then player2\isAI = 1
		; which has a combined cost of 12
		player2\isAI = -1 + KeyHit(22) + 2 * KeyHit(23) ; 7
		
		; we exit the loop once the seconds player's
		; AI status is correctly set
	Until player2\isAI >= 0 ; 2
	
	; initialise players positions and health
	player1\x = 40 ; 2
	player2\x = 40 ; 2
	
	player1\hp = 76 ; 2
	player2\hp = 76 ; 2
	
	Repeat ; 0 ; the main loop is free
		
		; clearing screen for new frame
		Cls ; 0 ; clearing is free
		
		; flushing keys to prevent unwanted behaviour
		FlushKeys() ; 1
		; using timer to syncronise to 20 FPS
		WaitTimer(timer) ; 2
		
		; drawing sky and background at different parallax
		TileImage(skyImage, backgroundParallax / 4) ; 4
		TileImage(backgroundImage, backgroundParallax) ; 3
		
		; calculating whether players are close enough to
		; hit each other;
		; another example of avoiding unneeded branching
		playersConnect = (player1\x + player2\x) < 10 ; 4
		
		; calculate if(and how much) players intersect, so
		; that they can be moved apart
		playersIntersect = (player1\x + player2\x) / 2 * ((player1\x + player2\x) < 0) ; 7
		
		; delegating identical code for both players to
		; functions saves instructions
		ControlPlayer(player1) ; 2
		ControlPlayer(player2) ; 2
		
		UpdatePlayer(player1, player2) ; 3
		UpdatePlayer(player2, player1) ; 3
		
		; draw players' health bars
		Rect(2, 2, player2\hp, 4) ; 5
		Rect(158 - player1\hp, 2, player1\hp, 4) ; 6
		
		; if either player has no health left,
		; exit the main loop to end the round
		If player1\hp <= 0 Or player2\hp <= 0 ; 5
			Exit ; 1
		End If
		
		; swap buffers, parameter FALSE, to avoid
		; 100% cpu for no reason
		Flip(False) ; 0 ; flip in mainloop is free
		
		; quit game on ESC
		If KeyHit(1) ; 0 ; ESC quit is free
			End ; 0
		End If
		
	Forever ; end main loop
	
Forever ; end repeating rounds loop

; this is the struct used to store information about
; the players
Type TPlayer
	Field x ; position
	Field y
	Field vx# ; velocity
	Field vy#
	Field hp ; health
	
	Field duck ; amoun player is ducking(if at all)
	Field isAI ; is AI?
	Field sign ; 1 for right, -1 for left player
	
	Field image ; the player's graphical representation
	
	Field keyUp ; keys
	Field keyDown
	Field keyCloseIn
	Field keyBlock
	Field keyPunch
	
	Field tryCloseIn ; intentions given by key pressing
	Field tryDuck    ; or AI
	Field tryBlock
	Field tryPunch
	Field tryJump
End Type

; this function creates a player with the given parameters
Function CreatePlayer.TPlayer(keyCloseIn, keyBlock, keyPunch, KeyD, keyU, image$, sign) ; 0
	Local player.TPlayer = New TPlayer ; 3
	player\keyCloseIn = keyCloseIn ; 2
	player\keyBlock = keyBlock ; 2
	player\keyPunch = keyPunch ; 2
	player\KeyDown = KeyD ; 2
	player\keyUp = keyU ; 2
	player\image = LoadAnimImage(image, 20, 30, 0, 4) ; 7
	player\sign = sign ; 2
	Return player
End Function

; this function accepts key presses to control the player
; or executes AI logic for AI player
Function ControlPlayer(player.TPlayer) ; 0
	; assign intentions from key presses
	player\tryCloseIn = KeyDown(player\keyCloseIn) ; 3
	player\tryBlock = KeyDown(player\keyBlock) ; 3
	player\tryPunch = KeyHit(player\keyPunch) ; 3
	player\tryDuck = KeyDown(player\KeyDown) ; 3
	player\tryJump = KeyHit(player\keyUp) ; 3
	
	If player\isAI ; 2
		; AI is always aggressive
		player\tryCloseIn = True ; 2
		; using the MOD operator results in repeating
		; behaviour; using different divisors for
		; different controls results in difficult to
		; predict behaviour, providing a good
		; challange
		player\tryDuck = (MilliSecs() Mod 3000) > 2200 ; 4
		player\tryJump = (MilliSecs() Mod 1400) > 1300 ; 4
		player\tryBlock = (MilliSecs() Mod 1700) > 1200 ; 4
		; the threshold below is the easiest way of
		; tweaking the difficulty of the AI,
		; with 0 being the easiest(it does not punch)
		; and 1 being impossible to beat
		; (punches every frame)
		player\tryPunch = Rnd(1) < 0.5 ; 4
	End If
End Function

; this function updates the player after its intentions
; have been set
Function UpdatePlayer(player.TPlayer, other.TPlayer) ; 0
	; make player duck 12 pixels when player tries to duck,
	; but also in flight
	player\duck = (player\tryDuck Or player\y < 50) * 12 ; 5
	; if player tries to jump and is on the ground,
	; they jump
	If player\tryJump And player\y = 50 ; 4
		player\vy = -6 ; 2
	End If
	; add gravity to vertical velocity
	player\vy = player\vy + 0.8 ; 3
	; add vertical velocity to player's position
	player\y = player\y + player\vy ; 3
	; if player is below ground, set to ground level
	If player\y > 50 ; 3
		player\y = 50 ; 2
	End If
	; this line, again avoiding branching controls the players
	; horizontal movement. It depends on whether the player
	; has the intention of closing in on the opponent,
	; whether they are ducking(no movement possible then),
	; and whether they are in the air(less friction and slowdown)
	player\vx = (player\vx - player\tryCloseIn * (player\duck = 0) * (player\tryBlock = 0)) * (0.7 + 0.29 * (player\y < 50)) ; 11
	; add horizontal velocity to position, also taking into
	; account players being too close and seperating them
	player\x = player\x + player\vx - playersIntersect ; 4
	
	; calculate if and how much the player is too close to
	; the screens edge
	Local off = (player\x - 50) / 2 * (player\x > 50) ; 6
	; move player back into view
	player\x = player\x - off ; 3
	; move background parallax to visualise players
	; moving across the scene
	backgroundParallax = backgroundParallax - off * player\sign ; 4
	
	; player punching logic here:
	; player can punch if he does not block
	If player\tryPunch - player\tryBlock > 0 ; 4
		; calculate height of punch in relation to other
		; other player's height
		Local h = player\y + player\duck + 10  - other\y ; 5
		; check if punch connected
		If playersConnect And h >= other\duck And h < 30 ; 6
			; substract 1 health from opponent
			; add 1 damage when player is in air
			; substract 1 damage if opponent is
			; blocking
			other\hp = other\hp - (1 + (player\y < 50) - other\tryBlock) ; 6
			; push opponent back, less if blocking
			other\vx = 3 - 2 * other\tryBlock ; 4
			; push player back if opponent is blocking
			player\vx = other\tryBlock ; 2
		End If
	End If
	; draw player at correct position, and with correct frame
	; frame is calculated using boolean opperations to avoid
	; branching again; also note that the punching and blocking
	; frames are the same
	DrawImage(player\image, 70 + player\sign * (player\x + 8), player\y, player\tryPunch Or player\tryBlock Or (2 * (player\duck <> 0))) ; 12
End Function

; That is all.
